# -*- coding: utf-8 -*-

import warnings

import sqlalchemy as sa
import sqlalchemy_utils
from sqlalchemy import create_engine
from sqlalchemy.engine import Engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.session import close_all_sessions

from sqlalchemy_i18n import (
    make_translatable,
    Translatable,
    translation_base,
    translation_manager
)
from sqlalchemy_i18n.manager import BaseTranslationMixin


@sa.event.listens_for(Engine, 'before_cursor_execute')
def count_sql_calls(conn, cursor, statement, parameters, context, executemany):
    conn.query_count += 1


make_translatable(options={'locales': ['en', 'fi']})


warnings.simplefilter('error', sa.exc.SAWarning)


sqlalchemy_utils.i18n.get_locale = lambda: 'en'


class DeclarativeTestCase(object):
    engine_uri = 'postgresql://postgres@localhost/sqlalchemy_i18n_test'
    locales = ['en', 'fi']
    create_tables = True
    configure_mappers = True

    def create_session(self):
        Session = sessionmaker(bind=self.connection)
        self.session = Session()

    def setup_method(self, method):
        self.engine = create_engine(self.engine_uri)
        # self.engine.echo = True
        self.connection = self.engine.connect()
        self.connection.query_count = 0
        self.Model = declarative_base()

        self.create_models()

        if self.configure_mappers:
            sa.orm.configure_mappers()
        if self.create_tables:
            self.Model.metadata.create_all(self.connection)
        self.create_session()

    def teardown_method(self, method):
        translation_manager.pending_classes = []
        close_all_sessions()
        self.Model.metadata.drop_all(self.connection)
        self.connection.close()
        self.engine.dispose()

    def create_models(self):
        class Article(self.Model, Translatable):
            __tablename__ = 'article'
            __translatable__ = {
                'locales': self.locales,
            }

            @hybrid_property
            def locale(self):
                return 'en'

            id = sa.Column(sa.Integer, autoincrement=True, primary_key=True)
            description = sa.Column(sa.UnicodeText)

            def __repr__(self):
                return 'Article(id=%r)' % self.id

        class ArticleTranslation(translation_base(Article)):
            __tablename__ = 'article_translation'

            name = sa.Column(sa.Unicode(255))
            content = sa.Column(sa.UnicodeText)

        self.Article = Article
        self.ArticleTranslation = ArticleTranslation

    def create_article(self):
        article = self.Article(name=u'Something', content=u'Something')
        self.session.add(article)
        self.session.commit()
        return article


class ClassicBase(object):
    def __init__(self, **kwargs):
        cls_ = type(self)
        for k in kwargs:
            if not hasattr(cls_, k):
                raise TypeError(
                    "%r is an invalid keyword argument for %s" %
                    (k, cls_.__name__))
            setattr(self, k, kwargs[k])

    def __repr__(self):
        "Return an ASCII representation of the entity."

        from sqlalchemy.orm.exc import DetachedInstanceError

        mapper = sa.orm.object_mapper(self)
        pkeyf = mapper.primary_key
        try:
            pkeyv = mapper.primary_key_from_instance(self)
        except DetachedInstanceError:
            keys = "detached-instance"
        else:
            keys = ', '.join(u'%s=%r' % (f.name, v)
                             for f, v in zip(pkeyf, pkeyv))
        return '%s(%s)' % (self.__class__.__name__, keys)


class ClassicTestCase(DeclarativeTestCase):
    def setup_method(self, method):
        self.metadata = sa.MetaData()

        self.engine = sa.create_engine(self.engine_uri)
        # self.engine.echo = True
        self.connection = self.engine.connect()
        self.connection.query_count = 0

        self.create_tables()
        self.create_models()
        self.create_mappers()

        if self.configure_mappers:
            sa.orm.configure_mappers()

        if self.create_tables:
            self.metadata.create_all(self.connection)

        self.create_session()

    def teardown_method(self, method):
        translation_manager.pending_classes = []
        close_all_sessions()
        self.metadata.drop_all(self.connection)
        self.connection.close()
        self.engine.dispose()

    def create_tables(self):
        self.article = sa.Table(
            'article', self.metadata,
            sa.Column(
                'id',
                sa.Integer,
                autoincrement=True,
                primary_key=True,
                nullable=False
            ),
            sa.Column('description', sa.UnicodeText)
        )
        self.article_translation = sa.Table(
            'article_translation',
            self.metadata,
            sa.Column(
                'id',
                sa.Integer,
                sa.ForeignKey(
                    'article',
                    ondelete="CASCADE"
                ),
                primary_key=True,
                nullable=False
            ),
            sa.Column(
                'locale',
                sa.types.CHAR(2),
                primary_key=True,
                nullable=False
            ),
            sa.Column('name', sa.Unicode(255)),
            sa.Column('content', sa.UnicodeText)
        )

    def create_models(self):
        class Article(ClassicBase, Translatable):
            __translatable__ = {
                'locales': self.locales,
            }
            __translated_columns__ = [
                sa.Column('name', sa.Unicode(255)),
                sa.Column('content', sa.UnicodeText),
            ]

            @hybrid_property
            def locale(self):
                return 'en'

        class ArticleTranslation(ClassicBase, BaseTranslationMixin):
            __parent_class__ = Article

        self.Article = Article
        self.ArticleTranslation = ArticleTranslation

    def create_mappers(self):
        sa.orm.mapper(self.Article, self.article)
        sa.orm.mapper(self.ArticleTranslation, self.article_translation)



# -*- coding: utf-8 -*-

import sqlalchemy as sa

from tests import ClassicTestCase, DeclarativeTestCase


class Suite(object):
    def test_joinedload_for_current_translation(self):
        article = (
            self.session.query(self.Article)
            .options(sa.orm.joinedload(self.Article.current_translation))
        ).first()
        query_count = self.connection.query_count
        article.name
        assert query_count == self.connection.query_count

    def test_contains_eager_for_current_translation(self):
        article = (
            self.session.query(self.Article)
            .join(self.Article.current_translation)
            .options(sa.orm.contains_eager(self.Article.current_translation))
        ).first()
        query_count = self.connection.query_count
        article.name
        assert query_count == self.connection.query_count

    def test_joinedload_for_single_translation(self):
        article = (
            self.session.query(self.Article)
            .options(sa.orm.joinedload(self.Article.translations['en']))
        ).first()
        query_count = self.connection.query_count
        article.name
        assert query_count == self.connection.query_count

    def test_joinedload_for_attr_accessor(self):
        article = (
            self.session.query(self.Article)
            .options(sa.orm.joinedload(self.Article.translations.en))
        ).first()
        query_count = self.connection.query_count
        article.name
        assert query_count == self.connection.query_count

    def test_joinedload_for_all_translations(self):
        article = (
            self.session.query(self.Article)
            .options(sa.orm.joinedload(self.Article.translations))
        ).first()
        query_count = self.connection.query_count
        article.name
        article.translations['fi'].name
        assert query_count == self.connection.query_count


class TestDeclarative(Suite, DeclarativeTestCase):
    def setup_method(self, method):
        DeclarativeTestCase.setup_method(self, method)
        article = self.Article(
            description=u'Some description',
            name=u'Some name'
        )
        self.session.add(article)
        self.session.commit()
        self.session.expunge_all()


class TestClassic(Suite, ClassicTestCase):
    def setup_method(self, method):
        ClassicTestCase.setup_method(self, method)
        article = self.Article(
            description=u'Some description',
            name=u'Some name'
        )
        self.session.add(article)
        self.session.commit()
        self.session.expunge_all()



# -*- coding: utf-8 -*-

import sqlalchemy as sa
from pytest import raises

from sqlalchemy_i18n import (
    ImproperlyConfigured,
    Translatable,
    translation_base
)
from tests import DeclarativeTestCase


class Suite(object):
    configure_mappers = False
    create_tables = False

    def test_raises_exception(self):
        with raises(ImproperlyConfigured):
            class ArticleTranslation(translation_base(self.Article)):
                __tablename__ = 'article_translation'

                name = sa.Column(sa.Unicode(255))
            self.Article()


class TestDeclarative(Suite, DeclarativeTestCase):
    def create_models(self):
        class Article(self.Model, Translatable):
            __tablename__ = 'article'
            __translatable__ = {
                'locales': ['fi', 'en'],
                'auto_create_locales': True,
            }

            id = sa.Column(sa.Integer, autoincrement=True, primary_key=True)

            name = sa.Column(sa.Unicode(255))

            locale = sa.Column(sa.Unicode(255), default=u'en')

            def get_locale(self):
                return 'en'

        self.Article = Article



# -*- coding: utf-8 -*-

import sqlalchemy as sa

from sqlalchemy_i18n import Translatable, translation_base
from sqlalchemy_i18n.manager import BaseTranslationMixin
from tests import ClassicBase
from tests.test_fallback_locale_as_callable import TestClassic as TCBase
from tests.test_fallback_locale_as_callable import TestDeclarative as TDBase


class TestDeclarative(TDBase):
    def create_models(self):
        class Article(self.Model, Translatable):
            __tablename__ = 'article'
            __translatable__ = {
                'locales': self.locales,
                'auto_create_locales': True,
            }

            id = sa.Column(sa.Integer, autoincrement=True, primary_key=True)
            locale = sa.Column(sa.Unicode(255), default=u'en')

            def get_locale(self):
                return 'en'

        class ArticleTranslation(translation_base(Article)):
            __tablename__ = 'article_translation'

            name = sa.Column(sa.Unicode(255))

        self.Article = Article


class TestClassic(TCBase):
    def create_models(self):
        class Article(ClassicBase, Translatable):
            __translatable__ = {
                'locales': self.locales,
                'auto_create_locales': True,
            }
            __translated_columns__ = [
                sa.Column('name', sa.Unicode(255)),
                sa.Column('content', sa.UnicodeText),
            ]

            def get_locale(self):
                return 'en'

        class ArticleTranslation(ClassicBase, BaseTranslationMixin):
            __parent_class__ = Article

        self.Article = Article
        self.ArticleTranslation = ArticleTranslation



# -*- coding: utf-8 -*-
import sqlalchemy as sa

from tests import ClassicTestCase, DeclarativeTestCase


class Suite(object):
    def test_current_translation_as_expression(self):
        query = (
            self.session.query(self.Article)
            .join(self.Article.current_translation)
        )
        assert (
            'JOIN article_translation ON article.id = article_translation.id'
            ' AND article_translation.locale = '
            in str(query)
        )

    def test_order_by_translation(self):
        self.session.add(self.Article(name=u'cbc'))
        self.session.add(self.Article(name=u'abc'))
        self.session.add(self.Article(name=u'bbc'))
        self.session.commit()

        current_translation = sa.orm.aliased(
            self.Article.current_translation.prop.mapper
        )
        articles = (
            self.session.query(self.Article)
            .join(current_translation, self.Article.current_translation)
            .options(sa.orm.contains_eager(
                self.Article.current_translation, alias=current_translation)
            )
            .order_by(current_translation.name)
        ).all()
        names = [a.name for a in articles]
        assert names == ['abc', 'bbc', 'cbc']

    def test_fallback_locale_as_expression(self):
        query = (
            self.session.query(self.Article)
            .join(self.Article.fallback_translation)
        )
        assert (
            'JOIN article_translation ON article.id = article_translation.id'
            ' AND article_translation.locale = '
            in str(query)
        )

    def test_querying(self):
        query = (
            self.session.query(self.Article)
        )
        assert str(query) == (
            'SELECT article.id AS article_id, article.description AS'
            ' article_description \nFROM article'
        )
        query = (
            self.session.query(self.Article.__translatable__['class'])
        )
        assert 'article_translation.id AS article_translation_id' in str(query)
        assert (
            'article_translation.locale AS article_translation_locale'
            in str(query)
        )
        assert (
            'article_translation.name AS article_translation_name'
            in str(query)
        )
        assert (
            'article_translation.content AS article_translation_content'
            in str(query)
        )
        assert 'FROM article_translation' in str(query)


class TestDeclarative(Suite, DeclarativeTestCase):
    pass


class TestClassic(Suite, ClassicTestCase):
    pass



# -*- coding: utf-8 -*-

from tests import ClassicTestCase, DeclarativeTestCase


class Suite(object):
    def test_as_object_property_with_force_locale(self):
        article = self.Article()
        article.name = u'Some article'
        article.content = u'Some content'
        self.session.add(article)
        self.session.commit()
        assert article.fallback_translation == article.translations['en']

    def test_as_class_property(self):
        assert self.Article.fallback_translation

    def test_setter(self):
        article = self.Article()
        article.fallback_translation = self.Article.__translatable__['class'](
            name=u'Something'
        )
        self.session.add(article)
        self.session.commit()
        assert article.name == u'Something'


class TestDeclarative(Suite, DeclarativeTestCase):
    pass


class TestClassic(Suite, ClassicTestCase):
    pass



# -*- coding: utf-8 -*-

from pytest import raises

from sqlalchemy_i18n import UnknownLocaleError
from tests import ClassicTestCase, DeclarativeTestCase


class Suite(object):
    def test_attribute_accessor_for_unknown_locale(self):
        with raises(UnknownLocaleError):
            assert self.Article.translations.some_unknown_locale

    def test_attribute_accessors(self):
        assert self.Article.translations.en
        assert self.Article.translations.fi


class TestDeclarative(Suite, DeclarativeTestCase):
    pass


class TestClassic(Suite, ClassicTestCase):
    pass



# -*- coding: utf-8 -*-

import sqlalchemy as sa

from sqlalchemy_i18n import Translatable, translation_base
from sqlalchemy_i18n.manager import BaseTranslationMixin
from tests import ClassicBase, ClassicTestCase, DeclarativeTestCase


class Suite(object):
    def test_translatable_dict_copied_to_each_child_class(self):
        assert (
            self.Article.__translatable__['class'] == self.ArticleTranslation
        )
        assert (
            self.TextItem.__translatable__['class'] == self.TextItemTranslation
        )


class TestDeclarative(Suite, DeclarativeTestCase):
    def create_models(self):
        class Base(self.Model, Translatable):
            __abstract__ = True
            __translatable__ = {
                'locales': ['en', 'fi'],
            }
            locale = 'en'

        class TextItem(Base):
            __tablename__ = 'text_item'

            id = sa.Column(sa.Integer, primary_key=True)

        class TextItemTranslation(translation_base(TextItem)):
            __tablename__ = 'text_item_translation'

            name = sa.Column(sa.Unicode(255))

        class Article(Base):
            __tablename__ = 'article'
            id = sa.Column(
                sa.Integer, primary_key=True
            )

        class ArticleTranslation(translation_base(Article)):
            __tablename__ = 'article_translation'

            name = sa.Column(sa.Unicode(255))

        self.TextItem = TextItem
        self.TextItemTranslation = TextItemTranslation
        self.Article = Article
        self.ArticleTranslation = ArticleTranslation


class TestClassic(Suite, ClassicTestCase):
    def create_tables(self):
        ClassicTestCase.create_tables(self)

        self.textitems = sa.Table(
            'textitems', self.metadata,
            sa.Column('id', sa.Integer,
                      autoincrement=True,
                      primary_key=True,
                      nullable=False))
        self.textitems_translation = sa.Table(
            'textitems_translation', self.metadata,
            sa.Column('id', sa.Integer, sa.ForeignKey('textitems'),
                      primary_key=True,
                      nullable=False),
            sa.Column('locale', sa.types.CHAR(2),
                      primary_key=True,
                      nullable=False),
            sa.Column('name', sa.UnicodeText))

    def create_models(self):
        class CommonBase(ClassicBase, Translatable):
            __translatable__ = {
                'locales': self.locales
            }
            locale = 'en'

        class TextItem(CommonBase):
            __translated_columns__ = [
                sa.Column('name', sa.UnicodeText),
            ]

        class TextItemTranslation(BaseTranslationMixin):
            __parent_class__ = TextItem

        class Article(CommonBase, Translatable):
            __translated_columns__ = [
                sa.Column('content', sa.UnicodeText),
            ]

        class ArticleTranslation(BaseTranslationMixin):
            __parent_class__ = Article

        self.TextItem = TextItem
        self.TextItemTranslation = TextItemTranslation
        self.Article = Article
        self.ArticleTranslation = ArticleTranslation

    def create_mappers(self):
        sa.orm.mapper(self.TextItem, self.textitems)
        sa.orm.mapper(self.TextItemTranslation, self.textitems_translation)
        ClassicTestCase.create_mappers(self)



# -*- coding: utf-8 -*-

import sqlalchemy as sa

from sqlalchemy_i18n import Translatable, translation_base
from tests import DeclarativeTestCase


class Suite(object):
    def test_translatable_dict_copied_to_each_child_class(self):
        assert issubclass(self.ArticleTranslation, self.TranslationBase)


class TestDeclarative(Suite, DeclarativeTestCase):
    def create_models(self):
        class Base(self.Model):
            __abstract__ = True
            __versioned__ = {}

        class Article(self.Model, Translatable):
            __tablename__ = 'article'
            id = sa.Column(
                sa.Integer, primary_key=True
            )

        class ArticleTranslation(
            translation_base(Article, base_class_factory=lambda cls: Base)
        ):
            __tablename__ = 'article_translation'

            name = sa.Column(sa.Unicode(255))

        self.TranslationBase = Base
        self.Article = Article
        self.ArticleTranslation = ArticleTranslation


# This is the common case in Classic mapping, no test needed



# -*- coding: utf-8 -*-

from pytest import raises

from sqlalchemy_i18n import UnknownLocaleError
from tests import ClassicTestCase, DeclarativeTestCase


class Suite(object):
    def test_proxy_contains(self):
        article = self.Article()
        article.translations['en']
        assert 'en' in article.translations

    def test_translation_mapping_attribute_getter(self):
        article = self.Article()
        article.translations.en.name = 'Something'
        assert article.name == 'Something'

    def test_attribute_accessor_for_unknown_locale(self):
        with raises(UnknownLocaleError):
            self.Article.translations.some_unknown_locale

    def test_proxy_not_contains(self):
        article = self.Article()
        assert 'unknown_locale' not in article.translations

    def test_items(self):
        article = self.create_article()
        assert isinstance(article.translations.items(), list)

    def test_iteritems(self):
        article = self.create_article()
        assert (
            article.translations.items() ==
            list(article.translations.iteritems())
        )

    def test_set_item(self):
        article = self.create_article()
        self.session.expunge_all()
        article = self.session.query(self.Article).first()
        locale_obj = self.Article.__translatable__['class'](
            name=u'Some other thing'
        )
        article.translations['en'] = locale_obj
        self.session.commit()
        self.session.expunge_all()
        article = self.session.query(self.Article).first()

        assert article.translations['en'].name == u'Some other thing'

    def test_repr(self):
        article = self.create_article()
        assert (
            repr(article.translations) ==
            'TranslationsMapping(Article(id=%d))' % article.id
        )

    def test_iter(self):
        article = self.create_article()
        assert len(list(article.translations)) == 2

    def test_values(self):
        article = self.create_article()
        values = list(article.translations.values())
        for value in values:
            assert isinstance(value, article.__translatable__['class'])
        assert len(values) == 2


class TestDeclarative(Suite, DeclarativeTestCase):
    pass


class TestClassic(Suite, ClassicTestCase):
    pass



# -*- coding: utf-8 -*-

from decimal import Decimal

from sqlalchemy_i18n import find_translations
from tests import ClassicTestCase, DeclarativeTestCase


class Suite(object):
    locales = ['en', 'fi', 'sv']

    def test_with_full_confidence(self):
        article = self.Article()
        article.translations['en'].name = u'Some article'
        article.translations['en'].content = u'Some content'
        article.translations['fi'].name = u'Joku artikkeli'
        article.translations['fi'].content = u'Jotain tekstii'
        self.session.add(article)
        self.session.commit()

        article = self.Article()
        article.translations['en'].name = u'Some article'
        self.session.add(article)
        self.session.commit()
        assert dict(find_translations(article, 'name', 'fi')) == {
            u'Joku artikkeli': Decimal('1.0')
        }

    def test_with_varying_confidence(self):
        article = self.Article()
        article.translations['en'].name = u'Some article'
        article.translations['en'].content = u'Some content'
        article.translations['fi'].name = u'Joku artikkeli'
        article.translations['fi'].content = u'Jotain tekstii'
        self.session.add(article)
        article = self.Article()
        article.translations['en'].name = u'Some article'
        article.translations['en'].content = u'Some content'
        article.translations['fi'].name = u'Joku toinen artikkeli'
        article.translations['fi'].content = u'Jotain muuta tekstii'
        self.session.add(article)

        self.session.commit()

        article = self.Article()
        article.translations['en'].name = u'Some article'
        self.session.add(article)
        self.session.commit()

        assert dict(find_translations(article, 'name', 'fi').all()) == {
            u'Joku artikkeli': Decimal('0.5'),
            u'Joku toinen artikkeli': Decimal('0.5')
        }


class TestDeclarative(Suite, DeclarativeTestCase):
    pass


class TestClassic(Suite, ClassicTestCase):
    pass



# -*- coding: utf-8 -*-

from tests import ClassicTestCase, DeclarativeTestCase


class Suite(object):
    def test_auto_creates_relations(self):
        article = self.Article()
        assert article.translations

    def test_translatable_attributes(self):
        article = self.Article()
        assert article.__translatable__['class']
        assert article.__translatable__['class'].__name__ == (
            'ArticleTranslation'
        )

    def test_relationship_consistency(self):
        article = self.Article()
        article.name = u'Some article'
        assert article.current_translation == article.translations['en']

    def test_property_delegators(self):
        article = self.Article()
        article.translations['en']

        assert not article.name
        article.current_translation.name = u'something'
        assert article.name == u'something'
        article.name = u'some other thing'
        assert article.current_translation.name == u'some other thing'
        assert article.translations['en'].name == u'some other thing'

    def test_commit_session(self):
        article = self.Article()
        article.name = u'Some article'
        article.content = u'Some content'
        self.session.add(article)
        self.session.commit()
        article = self.session.query(self.Article).get(1)
        assert article.name == u'Some article'
        assert article.content == u'Some content'

    def test_delete(self):
        article = self.Article()
        article.description = u'Some description'
        self.session.add(article)
        self.session.commit()
        article.name = u'Some article'
        article.content = u'Some content'
        self.session.delete(article)
        self.session.commit()


class TestDeclarative(Suite, DeclarativeTestCase):
    def test_translated_columns(self):
        article = self.Article()
        columns = article.__translatable__['class'].__table__.c
        assert 'name' in columns
        assert 'content' in columns
        assert 'description' not in columns

    def test_appends_locale_column_to_translation_table(self):
        table = self.Model.metadata.tables['article_translation']
        assert 'locale' in table.c


class TestClassic(Suite, ClassicTestCase):
    def test_translated_columns(self):
        columns = self.article_translation.c
        assert 'name' in columns
        assert 'content' in columns
        assert 'description' not in columns

    def test_appends_locale_column_to_translation_table(self):
        table = self.article_translation
        assert 'locale' in table.c



# -*- coding: utf-8 -*-

import sqlalchemy as sa

from sqlalchemy_i18n import Translatable, translation_base
from sqlalchemy_i18n.manager import BaseTranslationMixin
from tests import ClassicBase, ClassicTestCase, DeclarativeTestCase


class Suite(object):
    def test_hybrid_properties_support_callable_fallback_locales(self):
        article = self.Article(locale=u'en')
        article.name = u'Some article'
        assert article.name == u'Some article'

    def test_locale_fallback(self):
        article = self.Article(locale=u'en')
        article.name


class TestDeclarative(Suite, DeclarativeTestCase):
    def create_models(self):
        class Article(self.Model, Translatable):
            __tablename__ = 'article'
            __translatable__ = {
                'locales': self.locales,
                'auto_create_locales': True,
                'fallback_locale': lambda self: self.locale or 'en'
            }

            id = sa.Column(sa.Integer, autoincrement=True, primary_key=True)
            locale = sa.Column(sa.Unicode(255), default=u'en')

        class ArticleTranslation(translation_base(Article)):
            __tablename__ = 'article_translation'

            name = sa.Column(sa.Unicode(255))

        self.Article = Article


class TestClassic(Suite, ClassicTestCase):
    def create_tables(self):
        self.article = sa.Table(
            'article', self.metadata,
            sa.Column(
                'id',
                sa.Integer,
                autoincrement=True,
                primary_key=True,
                nullable=False
            ),
            sa.Column('locale', sa.types.CHAR(2), default='en'))
        self.article_translation = sa.Table(
            'article_translation', self.metadata,
            sa.Column(
                'id',
                sa.Integer,
                sa.ForeignKey('article'),
                primary_key=True,
                nullable=False
            ),
            sa.Column(
                'locale',
                sa.types.CHAR(2),
                primary_key=True,
                nullable=False
            ),
            sa.Column('name', sa.Unicode(255)),
            sa.Column('content', sa.UnicodeText)
        )

    def create_models(self):
        class Article(ClassicBase, Translatable):
            __translatable__ = {
                'locales': self.locales,
                'auto_create_locales': True,
                'fallback_locale': lambda self: self.locale or 'en'
            }
            __translated_columns__ = [
                sa.Column('name', sa.Unicode(255)),
                sa.Column('content', sa.UnicodeText),
            ]

        class ArticleTranslation(ClassicBase, BaseTranslationMixin):
            __parent_class__ = Article

        self.Article = Article
        self.ArticleTranslation = ArticleTranslation



import sqlalchemy as sa
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql.functions import GenericFunction


class current_locale(GenericFunction):
    type = sa.types.String()


@compiles(current_locale)
def compile_current_locale(element, compiler, **kw):
    # Lazy import get_locale so that it can be overridden
    from sqlalchemy_utils.i18n import get_locale

    return '%s' % compiler.process(
        sa.bindparam('current_locale', str(get_locale()))
    )



from copy import copy

import sqlalchemy as sa
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm.collections import attribute_mapped_collection
from sqlalchemy_utils.functions import get_primary_keys

from .comparators import TranslationComparator
from .exc import ImproperlyConfigured
from .expressions import current_locale
from .utils import get_fallback_locale, option


class HybridPropertyBuilder(object):
    def __init__(self, manager, translation_model):
        self.manager = manager
        self.translation_model = translation_model
        self.model = self.translation_model.__parent_class__

    def getter_factory(self, property_name):
        def attribute_getter(obj):
            value = getattr(obj.current_translation, property_name)
            if value:
                return value

            locale = get_fallback_locale(obj)
            return getattr(
                obj.translations[locale],
                property_name
            )

        return attribute_getter

    def setter_factory(self, property_name):
        """
        Return a hybrid property setter for given property name.

        :param property_name: Name of the property to generate a setter for
        """
        return (
            lambda obj, value:
            setattr(obj.current_translation, property_name, value)
        )

    def generate_hybrid(self, property_name):
        """
        Generate a SQLAlchemy hybrid property for given translation model
        property.

        :param property_name:
            Name of the translation model property to generate hybrid property
            for.
        """
        setattr(
            self.model,
            property_name,
            hybrid_property(
                fget=self.getter_factory(property_name),
                fset=self.setter_factory(property_name),
                expr=lambda cls: getattr(
                    cls.__translatable__['class'], property_name
                )
            )
        )

    def detect_collisions(self, property_name):
        """
        Detect possible naming collisions for given property name.

        :raises sqlalchemy_i18n.exc.ImproperlyConfigured: if the model already
            has a property with given name
        """
        mapper = sa.inspect(self.model)
        if mapper.has_property(property_name):
            raise ImproperlyConfigured(
                "Attribute name collision detected. Could not create "
                "hybrid property for translated attribute '%s'. "
                "An attribute with the same already exists in parent "
                "class '%s'." % (
                    property_name,
                    self.model.__name__
                )
            )

    def __call__(self):
        mapper = sa.orm.class_mapper(self.translation_model)
        for column in mapper.local_table.c:
            exclude = self.manager.option(
                self.model, 'exclude_hybrid_properties'
            )

            if column.key in exclude or column.primary_key:
                continue

            self.detect_collisions(column.key)
            self.generate_hybrid(column.key)


class RelationshipBuilder(object):
    def __init__(self, manager, translation_cls):
        self.manager = manager
        self.translation_cls = translation_cls
        self.parent_cls = self.translation_cls.__parent_class__

    @property
    def primary_key_conditions(self):
        conditions = []
        for key in get_primary_keys(self.parent_cls).keys():
            conditions.append(
                getattr(self.parent_cls, key) ==
                getattr(self.translation_cls, key)
            )
        return conditions

    def assign_single_translations(self):
        mapper = sa.orm.class_mapper(self.parent_cls)
        for locale in option(self.parent_cls, 'locales'):
            key = '_translation_%s' % locale
            if mapper.has_property(key):
                continue

            conditions = self.primary_key_conditions
            conditions.append(self.translation_cls.locale == locale)
            mapper.add_property(key, sa.orm.relationship(
                self.translation_cls,
                primaryjoin=sa.and_(*conditions),
                foreign_keys=list(
                    get_primary_keys(self.parent_cls).values()
                ),
                uselist=False,
                viewonly=True
            ))

    def assign_fallback_translation(self):
        """
        Assign the current translation relationship for translatable parent
        class.
        """
        mapper = sa.orm.class_mapper(self.parent_cls)
        if not mapper.has_property('_fallback_translation'):
            conditions = self.primary_key_conditions
            conditions.append(
                self.translation_cls.locale ==
                get_fallback_locale(self.parent_cls)
            )

            mapper.add_property('_fallback_translation', sa.orm.relationship(
                self.translation_cls,
                primaryjoin=sa.and_(*conditions),
                foreign_keys=list(
                    get_primary_keys(self.parent_cls).values()
                ),
                viewonly=True,
                uselist=False
            ))

    def assign_current_translation(self):
        """
        Assign the current translation relationship for translatable parent
        class.
        """
        mapper = sa.orm.class_mapper(self.parent_cls)
        if not mapper.has_property('_current_translation'):
            conditions = self.primary_key_conditions
            conditions.append(
                self.translation_cls.locale == current_locale()
            )

            mapper.add_property('_current_translation', sa.orm.relationship(
                self.translation_cls,
                primaryjoin=sa.and_(*conditions),
                foreign_keys=list(
                    get_primary_keys(self.parent_cls).values()
                ),
                viewonly=True,
                uselist=False
            ))

    def assign_translations(self):
        """
        Assigns translations relationship for translatable model. The assigned
        attribute is a relationship to all translation locales.
        """
        mapper = sa.orm.class_mapper(self.parent_cls)
        if not mapper.has_property('_translations'):
            mapper.add_property('_translations', sa.orm.relationship(
                self.translation_cls,
                **self.get_translations_relationship_args()
            ))

    def get_translations_relationship_args(self):
        foreign_keys = [
            getattr(self.translation_cls, column_key)
            for column_key in get_primary_keys(self.parent_cls).keys()
        ]

        relationship_args = copy(
            self.manager.option(
                self.parent_cls,
                'translations_relationship_args'
            )
        )
        defaults = dict(
            primaryjoin=sa.and_(*self.primary_key_conditions),
            foreign_keys=foreign_keys,
            collection_class=attribute_mapped_collection('locale'),
            comparator_factory=TranslationComparator,
            cascade='all, delete-orphan',
            passive_deletes=True,
        )
        for key, value in defaults.items():
            relationship_args.setdefault(key, value)
        return relationship_args

    def assign_translation_parent(self):
        mapper = sa.orm.class_mapper(self.translation_cls)
        if not mapper.has_property('translation_parent'):
            mapper.add_property('translation_parent', sa.orm.relationship(
                self.parent_cls,
                uselist=False,
                viewonly=True
            ))

    def __call__(self):
        self.assign_single_translations()
        self.assign_current_translation()
        self.assign_fallback_translation()
        self.assign_translations()
        self.assign_translation_parent()



from inspect import isclass

import sqlalchemy as sa


def get_current_locale(obj):
    from sqlalchemy_utils import i18n

    locale = i18n.get_locale()
    if locale:
        return str(locale)
    return obj.locale


def get_fallback_locale(obj):
    manager = obj.__translatable__['manager']

    locale = manager.option(obj, 'fallback_locale')
    if callable(locale):
        locale = locale(obj)
    return locale


def option(obj_or_class, option):
    class_ = obj_or_class if isclass(obj_or_class) else obj_or_class.__class__
    manager = class_.__translatable__['manager']
    return manager.option(class_, option)


def all_translated_columns(model):
    for column in sa.inspect(model.__translatable__['class']).columns:
        if not column.primary_key:
            yield column


def is_string(type_):
    return (
        isinstance(type_, sa.String) or
        (isclass(type_) and issubclass(type_, sa.String))
    )



from sqlalchemy.orm.relationships import RelationshipProperty

from .exc import UnknownLocaleError


class TranslationComparator(RelationshipProperty.Comparator):
    def __getitem__(self, key):
        return getattr(self._parentmapper.class_, '_translation_%s' % key)

    def __getattr__(self, locale):
        class_ = self._parentmapper.class_
        try:
            return getattr(class_, '_translation_%s' % locale)
        except AttributeError:
            raise UnknownLocaleError(locale, class_)



# -*- coding: utf-8 -*-

import sqlalchemy as sa

from sqlalchemy_i18n import Translatable, translation_base
from sqlalchemy_i18n.manager import BaseTranslationMixin
from tests import ClassicBase, ClassicTestCase, DeclarativeTestCase


class Suite(object):
    create_tables = False
    configure_mappers = False

    def test_does_not_generate_hybrid_properties(self):
        self.Article()


class TestDeclarative(Suite, DeclarativeTestCase):
    def create_models(self):
        class Article(self.Model, Translatable):
            __tablename__ = 'article'
            __translatable__ = {
                'locales': ['fi', 'en'],
                'auto_create_locales': True,
                'exclude_hybrid_properties': ['word_count']
            }

            id = sa.Column(sa.Integer, autoincrement=True, primary_key=True)

            word_count = sa.Column(sa.Integer)

            locale = sa.Column(sa.Unicode(255), default=u'en')

        class ArticleTranslation(translation_base(Article)):
            __tablename__ = 'article_translation'

            name = sa.Column(sa.Unicode(255))

            word_count = sa.Column(sa.Integer)

        self.Article = Article
        self.ArticleTranslation = ArticleTranslation


class TestClassic(Suite, ClassicTestCase):
    def create_tables(self):
        self.article = sa.Table(
            'article', self.metadata,
            sa.Column('id', sa.Integer,
                      autoincrement=True,
                      primary_key=True,
                      nullable=False),
            sa.Column('locale', sa.types.CHAR(2), default='en'),
            sa.Column('word_count', sa.Integer))
        self.article_translation = sa.Table(
            'article_translation', self.metadata,
            sa.Column('id', sa.Integer, sa.ForeignKey('article'),
                      primary_key=True,
                      nullable=False),
            sa.Column('locale', sa.types.CHAR(2),
                      primary_key=True,
                      nullable=False),
            sa.Column('name', sa.Unicode(255)),
            sa.Column('content', sa.UnicodeText),
            sa.Column('word_count', sa.Integer))

    def create_models(self):
        class Article(ClassicBase, Translatable):
            __translatable__ = {
                'locales': self.locales,
                'auto_create_locales': True,
                'exclude_hybrid_properties': ['word_count'],
            }
            __translated_columns__ = [
                sa.Column('name', sa.Unicode(255)),
                sa.Column('content', sa.UnicodeText),
            ]

        class ArticleTranslation(ClassicBase, BaseTranslationMixin):
            __parent_class__ = Article

        self.Article = Article
        self.ArticleTranslation = ArticleTranslation



import sqlalchemy as sa

from .builders import ImproperlyConfigured
from .manager import translation_base, translation_manager, TranslationManager
from .translatable import Translatable, UnknownLocaleError

__all__ = (
    translation_base,
    ImproperlyConfigured,
    Translatable,
    TranslationManager,
    translation_manager,
    UnknownLocaleError
)


__version__ = '1.1.0'


def make_translatable(
    mapper=sa.orm.mapper,
    session=sa.orm.session.Session,
    manager=translation_manager,
    options={}
):
    """
    Assigns translation listeners for given mapper and session.

    :param mapper:
        SQLAlchemy declarative class or mapper to apply translation listeners
        into.
    :param session:
        SQLAlchemy session class.
    :param manager:
        SQLAlchemy-i18n TranslationManager instance
    :param options:
        TranslationManager options
    """
    manager.options.update(options)

    sa.event.listen(
        mapper, 'instrument_class', manager.instrument_translation_classes
    )
    sa.event.listen(
        mapper, 'after_configured', manager.configure_translatable_classes
    )
    sa.event.listen(
        session, 'before_flush', manager.auto_create_translations
    )


def find_translations(obj, property_name, locale):
    class_ = obj.__class__
    session = sa.orm.object_session(obj)
    translation_class = class_.__translatable__['class']

    property_ = getattr(translation_class, property_name)

    subquery = (
        session.query(translation_class.id)
        .filter(
            sa.and_(
                property_ ==
                getattr(obj, property_name),
                translation_class.locale ==
                obj.locale
            )
        )
    )

    conditions = [
        translation_class.id.in_(subquery),
        translation_class.locale == locale,
        property_.isnot(None)
    ]

    total_count = (
        session.query(sa.func.cast(sa.func.count('1'), sa.Numeric))
        .filter(sa.and_(*conditions))
    )

    query = (
        session.query(
            property_.label('translation'),
            (sa.func.cast(sa.func.count('1'), sa.Numeric) / total_count)
            .label('confidence')
        )
        .filter(sa.and_(*conditions))
        .group_by(property_)
    )
    return query



# -*- coding: utf-8 -*-

import sqlalchemy as sa
from sqlalchemy.ext.hybrid import hybrid_property

from sqlalchemy_i18n import Translatable, translation_base
from sqlalchemy_i18n.manager import BaseTranslationMixin
from tests import ClassicBase, ClassicTestCase, DeclarativeTestCase


class Suite(object):
    def test_joinedload_for_current_translation(self):
        count = self.connection.query_count
        article = (
            self.session.query(self.Article)
            .options(sa.orm.joinedload(self.Article.current_translation))
        ).first()
        article.translations['en']

        assert self.connection.query_count == count + 1


class TestDeclarative(Suite, DeclarativeTestCase):
    def setup_method(self, method):
        DeclarativeTestCase.setup_method(self, method)
        self.ArticleTranslation = self.Article.__translatable__['class']

        article = self.Article(
            description=u'Some description',
            name=u'Some name'
        )
        self.session.add(article)
        self.session.commit()

    def create_models(self):
        class Article(self.Model, Translatable):
            __tablename__ = 'article'
            __translatable__ = {
                'locales': self.locales
            }

            @hybrid_property
            def locale(self):
                return self._locale or 'en'

            id = sa.Column(sa.Integer, autoincrement=True, primary_key=True)
            description = sa.Column(sa.UnicodeText)
            _locale = sa.Column(sa.Unicode(10), default=u'en')

        class ArticleTranslation(translation_base(Article)):
            __tablename__ = 'article_translation'

            name = sa.Column(sa.Unicode(255))
            content = sa.Column(sa.UnicodeText)

        self.Article = Article


class TestClassic(Suite, ClassicTestCase):
    def setup_method(self, method):
        ClassicTestCase.setup_method(self, method)

        article = self.Article(description=u'Some description',
                               name=u'Some name')
        self.session.add(article)
        self.session.commit()

    def create_tables(self):
        self.article = sa.Table(
            'article', self.metadata,
            sa.Column('id', sa.Integer,
                      autoincrement=True,
                      primary_key=True,
                      nullable=False),
            sa.Column('description', sa.UnicodeText),
            sa.Column('_locale', sa.types.CHAR(2), default=u'en'))
        self.article_translation = sa.Table(
            'article_translation', self.metadata,
            sa.Column('id', sa.Integer, sa.ForeignKey('article'),
                      primary_key=True,
                      nullable=False),
            sa.Column('locale', sa.types.CHAR(2),
                      primary_key=True,
                      nullable=False),
            sa.Column('name', sa.Unicode(255)),
            sa.Column('content', sa.UnicodeText))

    def create_models(self):
        class Article(ClassicBase, Translatable):
            __translatable__ = {
                'locales': self.locales
            }
            __translated_columns__ = [
                sa.Column('name', sa.Unicode(255)),
                sa.Column('content', sa.UnicodeText),
            ]

            @hybrid_property
            def locale(self):
                return self._locale or 'en'

        class ArticleTranslation(ClassicBase, BaseTranslationMixin):
            __parent_class__ = Article

        self.Article = Article
        self.ArticleTranslation = ArticleTranslation



# -*- coding: utf-8 -*-

import sqlalchemy as sa

from sqlalchemy_i18n import Translatable, translation_base
from tests import DeclarativeTestCase


class TestTranslationBaseCustomization(DeclarativeTestCase):
    def create_models(self):
        class Article(self.Model, Translatable):
            __tablename__ = 'article'
            __translatable__ = {
                'locales': ['fi', 'en'],
                'auto_create_locales': True,
                'translations_relationship_args': {
                    'collection_class': list,
                }
            }

            id = sa.Column(sa.Integer, primary_key=True)

            locale = 'en'

        TranslationBase = translation_base(
            Article,
            foreign_key_args={'onupdate': 'SET NULL'}
        )

        class ArticleTranslation(TranslationBase):
            __tablename__ = 'article_translation'

            name = sa.Column(sa.Unicode(255))

        self.Article = Article
        self.ArticleTranslation = ArticleTranslation

    def test_customize_foreign_key_args(self):
        assert (
            self.ArticleTranslation.__table__.foreign_keys.pop().onupdate ==
            'SET NULL'
        )

    def test_customize_relationships_args(self):
        assert self.Article._translations.property.collection_class == list



# -*- coding: utf-8 -*-

import sqlalchemy as sa

from sqlalchemy_i18n import Translatable, translation_base
from sqlalchemy_i18n.manager import BaseTranslationMixin
from tests import ClassicBase, ClassicTestCase, DeclarativeTestCase


class Suite(object):
    def test_relationships(self):
        article = self.Article(id1=1, id2=1)
        article.name = u'Some article'
        self.session.add(article)
        self.session.commit()

        assert article.name == u'Some article'


class TestDeclarative(Suite, DeclarativeTestCase):
    def create_models(self):
        class Article(self.Model, Translatable):
            __tablename__ = 'article'
            __translatable__ = {
                'locales': ['fi', 'en'],
                'auto_create_locales': True,
            }

            id1 = sa.Column(sa.Integer, primary_key=True)

            id2 = sa.Column(sa.Integer, primary_key=True)

            locale = 'en'

        class ArticleTranslation(translation_base(Article)):
            __tablename__ = 'article_translation'

            name = sa.Column(sa.Unicode(255))

        self.Article = Article
        self.ArticleTranslation = ArticleTranslation


class TestClassic(Suite, ClassicTestCase):
    def create_tables(self):
        self.article = sa.Table(
            'article',
            self.metadata,
            sa.Column(
                'id1',
                sa.Integer,
                primary_key=True,
                nullable=False
            ),
            sa.Column(
                'id2',
                sa.Integer,
                primary_key=True,
                nullable=False
            )
        )
        self.article_translation = sa.Table(
            'article_translation',
            self.metadata,
            sa.Column(
                'id1',
                sa.Integer,
                primary_key=True,
                nullable=False
            ),
            sa.Column(
                'id2',
                sa.Integer,
                primary_key=True,
                nullable=False
            ),
            sa.Column(
                'locale',
                sa.types.CHAR(2),
                primary_key=True,
                nullable=False
            ),
            sa.Column('name', sa.Unicode(255)),
            sa.Column('content', sa.UnicodeText),
            sa.ForeignKeyConstraint(
                ['id1', 'id2'],
                ['article.id1', 'article.id2']
            )
        )

    def create_models(self):
        class Article(ClassicBase, Translatable):
            __translatable__ = {
                'locales': self.locales
            }
            __translated_columns__ = [
                sa.Column('content', sa.UnicodeText),
            ]
            locale = 'en'

        class ArticleTranslation(ClassicBase, BaseTranslationMixin):
            __parent_class__ = Article

        self.Article = Article
        self.ArticleTranslation = ArticleTranslation



import sqlalchemy as sa
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm.util import has_identity

from .exc import UnknownLocaleError
from .utils import get_current_locale, get_fallback_locale


class Translatable(object):
    __translatable__ = {
        'fallback_locale': 'en'
    }

    @hybrid_property
    def locale(self):
        raise NotImplementedError(
            'Your translatable model needs to define locale property.'
        )

    # Current translation getters and setters
    @hybrid_property
    def current_translation(self):
        return self.translations[get_current_locale(self)]

    @current_translation.setter
    def current_translation(self, obj):
        self.translations[get_current_locale(self)] = obj

    @current_translation.expression
    def current_translation(cls):
        return cls._current_translation

    # Fallback translation getters and setters
    @hybrid_property
    def fallback_translation(self):
        return self.translations[get_fallback_locale(self)]

    @fallback_translation.setter
    def fallback_translation(self, obj):
        self.translations[get_fallback_locale(self)] = obj

    @fallback_translation.expression
    def fallback_translation(cls):
        return cls._fallback_translation

    # Translations getters and setters
    @hybrid_property
    def translations(self):
        if not hasattr(self, '_translations_mapping'):
            self._translations_mapping = TranslationsMapping(self)
        return self._translations_mapping

    @translations.setter
    def translations(self, translations_mapping):
        self._translations = translations_mapping

    @translations.expression
    def translations(self):
        return self._translations


@sa.event.listens_for(sa.orm.mapper, 'expire')
def receive_expire(target, attrs):
    if isinstance(target, Translatable):
        try:
            del target._translations_mapping
        except AttributeError:
            pass


class TranslationsMapping(object):
    def __init__(self, obj):
        self.obj = obj
        self.manager = self.obj.__translatable__['manager']

    def __contains__(self, locale):
        return locale in self.manager.option(self.obj, 'locales')

    def fetch(self, locale):
        session = sa.orm.object_session(self.obj)
        # If the object has no identity and its not in session or if the object
        # has _translations relationship loaded get the locale object from the
        # relationship.
        if '_translations' not in sa.inspect(self.obj).unloaded or (
            not session or not has_identity(self.obj)
        ):
            return self.obj._translations.get(locale)

        return session.query(self.obj.__translatable__['class']).get(
            sa.inspect(self.obj).identity + (locale, )
        )

    def __getitem__(self, locale):
        if locale in self:
            locale_obj = self.fetch(locale)
            if locale_obj is not None:
                return locale_obj

            class_ = self.obj.__translatable__['class']
            locale_obj = class_(
                translation_parent=self.obj,
                locale=locale
            )
            self.obj._translations[locale] = locale_obj
            return locale_obj
        raise UnknownLocaleError(locale, self.obj)

    def __getattr__(self, locale):
        return self.__getitem__(locale)

    # Added for py2.6 compatibility
    def __length_hint__(self):
        return len(self)

    def __len__(self):
        return len(self.manager.option(self.obj, 'locales'))

    @property
    def all(self):
        return list(self.values())

    def values(self):
        for locale in self.manager.option(self.obj, 'locales'):
            yield self[locale]

    def __setitem__(self, locale, translation_obj):
        if locale in self:
            translation_obj.translation_parent = self.obj
            translation_obj.locale = locale
            self.obj._translations[locale] = translation_obj

    def __repr__(self):
        return 'TranslationsMapping(%r)' % self.obj

    def items(self):
        return [
            (locale, self[locale])
            for locale in self.manager.option(self.obj, 'locales')
        ]

    def iteritems(self):
        for locale in self.manager.option(self.obj, 'locales'):
            yield locale, self[locale]

    def __iter__(self):
        for locale in self.manager.option(self.obj, 'locales'):
            yield locale, self[locale]



# -*- coding: utf-8 -*-

from sqlalchemy_i18n import translation_manager
from tests import ClassicTestCase, DeclarativeTestCase


class Suite(object):
    def test_auto_creates_translation_objects(self):
        article = self.Article(name=u'Some article')
        self.session.add(article)
        self.session.commit()

        assert 'en' in article.translations
        assert article.translations['en']
        assert 'fi' in article.translations
        assert article.translations['fi']


class TestDeclarative(Suite, DeclarativeTestCase):
    def setup_method(self, method):
        DeclarativeTestCase.setup_method(self, method)
        translation_manager.options['auto_created_locales'] = [u'fi', u'en']

    def teardown_method(self, method):
        DeclarativeTestCase.teardown_method(self, method)
        translation_manager.options['auto_created_locales'] = []


class TestClassic(Suite, ClassicTestCase):
    def setup_method(self, method):
        ClassicTestCase.setup_method(self, method)
        translation_manager.options['auto_created_locales'] = [u'fi', u'en']

    def teardown_method(self, method):
        ClassicTestCase.teardown_method(self, method)
        translation_manager.options['auto_created_locales'] = []



from copy import copy

import sqlalchemy as sa
from sqlalchemy.ext.declarative import declared_attr, has_inherited_table
from sqlalchemy_utils.functions import get_declarative_base, get_primary_keys

from .builders import HybridPropertyBuilder, RelationshipBuilder
from .utils import all_translated_columns, is_string


class BaseTranslationMixin(object):
    pass


def translation_base(
    parent_cls,
    base_class_factory=None,
    foreign_key_args=None
):
    if base_class_factory is None:
        base_class_factory = get_declarative_base

    if foreign_key_args is None:
        foreign_key_args = {}
        foreign_key_args.setdefault('ondelete', 'CASCADE')

    class TranslationMixin(
        base_class_factory(parent_cls),
        BaseTranslationMixin
    ):
        __abstract__ = True
        __parent_class__ = parent_cls

        @declared_attr
        def __table_args__(cls):
            if has_inherited_table(cls):
                return tuple()
            else:
                names = list(get_primary_keys(parent_cls).keys())

                return (
                    sa.schema.ForeignKeyConstraint(
                        names,
                        [
                            '%s.%s' % (parent_cls.__tablename__, name)
                            for name in names
                        ],
                        **foreign_key_args
                    ),
                )

    for column in parent_cls.__table__.c:
        if column.primary_key:
            column_copy = column._copy()
            column_copy.autoincrement = False
            setattr(
                TranslationMixin,
                column.key,
                column_copy
            )

    TranslationMixin.locale = sa.Column(
        sa.String(10), primary_key=True
    )

    return TranslationMixin


class TranslationManager(object):
    def __init__(self):
        self.class_map = {}
        self.pending_classes = []
        self.options = {
            'locales': [],
            'auto_create_locales': True,
            'fallback_locale': 'en',
            'exclude_hybrid_properties': [],
            'translations_relationship_args': {}
        }

    def instrument_translation_classes(self, mapper, cls):
        """
        SQLAlchemy class instrumentation listener that adds all translation
        classes to pending classes list. These classes are later on processed
        by configure_translatable_classes listener.

        :param mapper: SQLAlchemy mapper
        :param cls: SQLAlchemy declarative class
        """
        if issubclass(cls, BaseTranslationMixin):
            self.pending_classes.append(cls)

    def configure_translatable_classes(self):
        """
        This SQLAlchemy after_configured listener configures all translation
        classes, builds hybrid properties for translation parent classes and
        finally builds relationships between translation and parent classes.
        """
        for cls in self.pending_classes:
            self.class_map[cls.__parent_class__] = cls
            parent_cls = cls.__parent_class__
            # Args need to be copied to avoid scenarios where many child
            # classes inherit the __translatable__ dict from parent class and
            # then have the same reference to same dict.
            parent_cls.__translatable__ = copy(parent_cls.__translatable__)
            parent_cls.__translatable__['manager'] = self
            parent_cls.__translatable__['class'] = cls

            HybridPropertyBuilder(self, cls)()
            RelationshipBuilder(self, cls)()

        self.pending_classes = []

    def option(self, model, name):
        """
        Returns the option value for given model. If the option is not found
        from given model falls back to default values of this manager object.
        If the option is not found from this manager object either this method
        throws a KeyError.

        :param model: SQLAlchemy declarative object
        :param name: name of the translation option
        """
        try:
            return model.__translatable__[name]
        except (AttributeError, KeyError):
            return self.options[name]

    def set_not_nullables_to_empty_strings(self, locale, obj):
        for column in all_translated_columns(obj.__class__):
            if (
                not column.nullable and
                is_string(column.type) and
                column.default is None and
                column.server_default is None and
                getattr(obj.translations[locale], column.name) is None
            ):
                setattr(obj.translations[locale], column.name, u'')

    def create_missing_locales(self, obj):
        """
        Creates empty locale objects for given SQLAlchemy declarative object.

        :param model: SQLAlchemy declarative model object
        """
        session = sa.orm.session.object_session(obj)
        for locale in self.option(obj, 'locales'):
            if obj.translations[locale] is None:
                class_ = obj.__translatable__['class']
                obj.translations[locale] = class_(
                    translation_parent=obj,
                    locale=locale
                )
            self.set_not_nullables_to_empty_strings(locale, obj)
            session.add(obj)

    def auto_create_translations(self, session, flush_context, instances):
        if self.options['auto_create_locales']:
            for obj in session.new:
                if hasattr(obj, '__translatable__'):
                    self.create_missing_locales(obj)


translation_manager = TranslationManager()



from .utils import option


class ImproperlyConfigured(Exception):
    pass


class UnknownLocaleError(Exception):
    def __init__(self, locale, obj):
        Exception.__init__(
            self,
            'Unknown locale %s given for instance of class %r. '
            'Locale is not one of %r' % (
                locale, obj.__class__, list(option(obj, 'locales'))
            )
        )



"""
SQLAlchemy-i18n
---------------

Internationalization extension for SQLAlchemy models.
"""
import os
import re
import sys
from setuptools import setup


PY3 = sys.version_info[0] == 3
HERE = os.path.dirname(os.path.abspath(__file__))


def get_version():
    filename = os.path.join(HERE, 'sqlalchemy_i18n', '__init__.py')
    with open(filename) as f:
        contents = f.read()
    pattern = r"^__version__ = '(.*?)'$"
    return re.search(pattern, contents, re.MULTILINE).group(1)


extras_require = {
    'test': [
        'pytest>=2.2.3',
        'Pygments>=1.2',
        'Jinja2>=2.3',
        'docutils>=0.10',
        'flexmock>=0.9.7',
        'psycopg2>=2.4.6',
        'flake8>=2.4.0',
        'isort>=3.9.6'
    ]
}


setup(
    name='SQLAlchemy-i18n',
    version=get_version(),
    url='https://github.com/kvesteri/sqlalchemy-i18n',
    license='BSD',
    author='Konsta Vesterinen',
    author_email='konsta@fastmonkeys.com',
    description='Internationalization extension for SQLAlchemy models.',
    long_description=__doc__,
    packages=['sqlalchemy_i18n'],
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    install_requires=[
        'SQLAlchemy>=1.3',
        'SQLAlchemy-Utils>=0.37.3',
        'six>=1.4.1'
    ],
    python_requires='>=3.6',
    extras_require=extras_require,
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)



# -*- coding: utf-8 -*-

import sqlalchemy as sa

from sqlalchemy_i18n import Translatable, translation_base
from sqlalchemy_i18n.manager import BaseTranslationMixin
from tests import ClassicBase, ClassicTestCase, DeclarativeTestCase


class Suite(object):
    def test_nullify_relation(self):
        self.CategoryTranslation = self.Category.__translatable__['class']
        article = self.Article()
        article.content = u'Some content'
        article.categories.append(self.Category())
        self.session.commit()
        article.categories = []
        self.session.commit()
        assert not self.session.query(self.CategoryTranslation).count()


class TestDeclarative(Suite, DeclarativeTestCase):
    def create_models(self):
        class Article(self.Model):
            __tablename__ = 'article'
            id = sa.Column(sa.Integer, autoincrement=True, primary_key=True)
            content = sa.Column(sa.UnicodeText)

        class Category(self.Model, Translatable):
            __tablename__ = 'category'
            id = sa.Column(sa.Integer, autoincrement=True, primary_key=True)
            article_id = sa.Column(sa.Integer, sa.ForeignKey(Article.id))
            article = sa.orm.relationship(
                Article,
                backref=sa.orm.backref(
                    'categories',
                    passive_deletes=True,
                    cascade='all, delete-orphan'
                )
            )

        class CategoryTranslation(translation_base(Category)):
            __tablename__ = 'article_translation'

            name = sa.Column(sa.Unicode(255))

        self.Article = Article
        self.Category = Category


class TestClassic(Suite, ClassicTestCase):
    def create_tables(self):
        self.article = sa.Table(
            'article', self.metadata,
            sa.Column('id', sa.Integer,
                      autoincrement=True,
                      primary_key=True,
                      nullable=False),
            sa.Column('content', sa.UnicodeText))
        self.category = sa.Table(
            'category', self.metadata,
            sa.Column('id', sa.Integer,
                      autoincrement=True,
                      primary_key=True,
                      nullable=False),
            sa.Column('article_id', sa.Integer, sa.ForeignKey('article.id')))
        self.category_translation = sa.Table(
            'category_translation', self.metadata,
            sa.Column('id', sa.Integer, sa.ForeignKey('category'),
                      primary_key=True,
                      nullable=False),
            sa.Column('locale', sa.types.CHAR(2),
                      primary_key=True,
                      nullable=False),
            sa.Column('name', sa.Unicode(255)))

    def create_models(self):
        class Article(ClassicBase):
            __translatable__ = {
                'locales': self.locales,
                'default_locale': 'en',
            }
            __translated_columns__ = [
                sa.Column('name', sa.Unicode(255)),
                sa.Column('content', sa.UnicodeText),
            ]

        class Category(ClassicBase, Translatable):
            __translated_columns__ = [
                sa.Column('name', sa.Unicode(255)),
            ]

        class CategoryTranslation(ClassicBase, BaseTranslationMixin):
            __parent_class__ = Category

        self.Article = Article
        self.Category = Category
        self.CategoryTranslation = CategoryTranslation

    def create_mappers(self):
        sa.orm.mapper(self.Article, self.article)
        sa.orm.mapper(self.Category, self.category, properties={
            'article': sa.orm.relationship(
                self.Article,
                backref=sa.orm.backref('categories',
                                       passive_deletes=True,
                                       cascade='all, delete-orphan'))
        })
        sa.orm.mapper(self.CategoryTranslation, self.category_translation)

    def test_nullify_relation(self):
        article = self.Article()
        article.content = u'Some content'
        article.categories.append(self.Category())
        self.session.commit()
        article.categories = []
        self.session.commit()
        assert not self.session.query(self.CategoryTranslation).count()



# -*- coding: utf-8 -*-

import sqlalchemy as sa

from sqlalchemy_i18n import Translatable, translation_base
from sqlalchemy_i18n.manager import BaseTranslationMixin
from tests import ClassicBase, ClassicTestCase, DeclarativeTestCase


class Suite(object):
    def test_auto_creates_relations(self):
        # textitem = self.TextItem()
        # assert textitem.translations
        # assert textitem._translations
        article = self.Article()
        assert article.translations

    def test_auto_creates_current_translation(self):
        # textitem = self.TextItem()
        # assert textitem.current_translation
        article = self.Article()
        assert article.current_translation

    def test_translatable_attributes(self):
        textitem = self.TextItem()
        class_ = textitem.__translatable__['class']
        assert class_.__name__ == 'TextItemTranslation'

    def test_inherits_child_properties(self):
        assert self.Article.caption

    def test_translated_columns(self):
        article = self.Article()
        columns = sa.inspect(article.__translatable__['class']).columns
        assert 'caption' in columns
        assert 'name' in columns
        assert 'content' in columns
        assert 'description' not in columns

    def test_property_delegators(self):
        article = self.Article()
        article.translations['en']

        assert not article.name
        article.current_translation.name = u'something'
        assert article.name == u'something'
        article.name = u'some other thing'
        assert article.current_translation.name == u'some other thing'
        assert article.translations['en'].name == u'some other thing'
        article.caption = u'some caption'
        assert article.current_translation.caption == u'some caption'


class TestDeclarative(Suite, DeclarativeTestCase):
    def create_models(self):
        class TextItem(self.Model, Translatable):
            __tablename__ = 'text_item'
            __translatable__ = {
                'locales': ['en', 'fi'],
            }
            locale = 'en'

            id = sa.Column(sa.Integer, autoincrement=True, primary_key=True)
            description = sa.Column(sa.UnicodeText)
            discriminator = sa.Column(
                sa.Unicode(100)
            )
            __mapper_args__ = {
                'polymorphic_on': discriminator,
            }

        class TextItemTranslation(translation_base(TextItem)):
            __tablename__ = 'text_item_translation'
            name = sa.Column(sa.Unicode(255))

            content = sa.Column(sa.UnicodeText)

            caption = sa.Column(sa.UnicodeText)

        class Article(TextItem):
            __tablename__ = 'article'
            id = sa.Column(
                sa.Integer, sa.ForeignKey(TextItem.id), primary_key=True
            )
            __mapper_args__ = {'polymorphic_identity': u'article'}
            category = sa.Unicode(255)

        self.TextItem = TextItem
        self.Article = Article

    def test_table_name(self):
        article = self.Article()
        table = article.__translatable__['class'].__table__
        assert table.name == 'text_item_translation'

    def test_inherits_parent_table(self):
        table = self.Article.__translatable__['class'].__table__
        assert table == self.TextItem.__translatable__['class'].__table__


class TestClassic(Suite, ClassicTestCase):
    def create_tables(self):
        self.text_item = sa.Table(
            'text_item', self.metadata,
            sa.Column('id', sa.Integer,
                      autoincrement=True,
                      primary_key=True,
                      nullable=False),
            sa.Column('description', sa.UnicodeText),
            sa.Column('discriminator', sa.Unicode(255)))
        self.text_item_translation = sa.Table(
            'text_item_translation', self.metadata,
            sa.Column('id', sa.Integer, sa.ForeignKey('text_item.id',
                                                      ondelete="CASCADE"),
                      primary_key=True,
                      nullable=False),
            sa.Column('locale', sa.types.CHAR(2),
                      primary_key=True,
                      nullable=False),
            sa.Column('name', sa.Unicode(255), nullable=False),
            sa.Column('content', sa.UnicodeText, nullable=False),
            sa.Column('caption', sa.UnicodeText, nullable=False))
        self.article = sa.Table(
            'article', self.metadata,
            sa.Column('id', sa.Integer, sa.ForeignKey('text_item.id',
                                                      ondelete="CASCADE"),
                      autoincrement=True,
                      primary_key=True,
                      nullable=False),
            sa.Column('category', sa.Unicode(255)))

    def create_models(self):
        class TextItem(ClassicBase, Translatable):
            __translatable__ = {
                'locales': self.locales,
            }
            __translated_columns__ = [
                sa.Column('name', sa.Unicode(255)),
                sa.Column('content', sa.UnicodeText),
                sa.Column('content2', sa.UnicodeText),
            ]
            locale = 'en'

        class TextItemTranslation(ClassicBase, BaseTranslationMixin):
            __parent_class__ = TextItem

        class Article(TextItem):
            pass

        self.TextItem = TextItem
        self.TextItemTranslation = TextItemTranslation
        self.Article = Article

    def create_mappers(self):
        sa.orm.mapper(self.TextItem, self.text_item,
                      polymorphic_on=self.text_item.c.discriminator)
        sa.orm.mapper(self.TextItemTranslation, self.text_item_translation)
        sa.orm.mapper(self.Article, self.article,
                      inherits=self.TextItem,
                      polymorphic_identity=u'article')

    def test_table_name(self):
        article_l10n_mapper = sa.orm.class_mapper(
            self.Article.__translatable__['class']
        )
        table = article_l10n_mapper.local_table
        assert table.name == 'text_item_translation'

    def test_inherits_parent_table(self):
        article_l10n_mapper = sa.orm.class_mapper(
            self.Article.__translatable__['class']
        )
        textitem_l10n_mapper = sa.orm.class_mapper(
            self.TextItem.__translatable__['class']
        )
        article_l10n_table = article_l10n_mapper.local_table
        textitem_l10n_table = textitem_l10n_mapper.local_table
        assert article_l10n_table == textitem_l10n_table



# -*- coding: utf-8 -*-

import sqlalchemy as sa
from sqlalchemy.ext.hybrid import hybrid_property

from sqlalchemy_i18n import Translatable, translation_base, translation_manager
from sqlalchemy_i18n.manager import BaseTranslationMixin
from tests import ClassicBase, ClassicTestCase, DeclarativeTestCase


class Suite(object):
    def test_auto_sets_nullables_as_empty_strings(self):
        article = self.ExtendedArticle(
            name=u'Some article',
            content2=u'Some content'
        )
        self.session.add(article)
        self.session.commit()

        assert article.translations['en'].name == u'Some article'
        assert article.translations['fi'].name == u''


class TestDeclarative(Suite, DeclarativeTestCase):
    def setup_method(self, method):
        DeclarativeTestCase.setup_method(self, method)
        translation_manager.options['auto_created_locales'] = [u'fi', u'en']

    def teardown_method(self, method):
        DeclarativeTestCase.teardown_method(self, method)
        translation_manager.options['auto_created_locales'] = []

    def create_models(self):
        class Article(self.Model, Translatable):
            __tablename__ = 'article'
            __translatable__ = {
                'auto_created_locales': True,
                'locales': ['en', 'fi']
            }

            @hybrid_property
            def locale(self):
                return 'en'

            id = sa.Column(sa.Integer, autoincrement=True, primary_key=True)
            description = sa.Column(sa.UnicodeText)
            discriminator = sa.Column(sa.Unicode(255))

            __mapper_args__ = {
                'polymorphic_on': discriminator,
            }

        class ExtendedArticle(Article):
            __tablename__ = 'extended_article'
            __mapper_args__ = {'polymorphic_identity': u'extended'}
            id = sa.Column(
                sa.Integer, sa.ForeignKey(Article.id), primary_key=True
            )

        class ArticleTranslation(translation_base(Article)):
            __tablename__ = 'article_translation'

            name = sa.Column(sa.Unicode(255), nullable=False)

            content = sa.Column(sa.UnicodeText)

            content2 = sa.Column(sa.UnicodeText)

        self.Article = Article
        self.ExtendedArticle = ExtendedArticle


class TestClassic(Suite, ClassicTestCase):
    def setup_method(self, method):
        ClassicTestCase.setup_method(self, method)
        translation_manager.options['auto_created_locales'] = [u'fi', u'en']

    def teardown_method(self, method):
        ClassicTestCase.teardown_method(self, method)
        translation_manager.options['auto_created_locales'] = []

    def create_tables(self):
        self.article = sa.Table(
            'article', self.metadata,
            sa.Column('id', sa.Integer,
                      autoincrement=True,
                      primary_key=True,
                      nullable=False),
            sa.Column('description', sa.UnicodeText),
            sa.Column('discriminator', sa.Unicode(255)))
        self.extended_article = sa.Table(
            'extended_article', self.metadata,
            sa.Column('id', sa.Integer, sa.ForeignKey('article.id'),
                      primary_key=True))
        self.article_translation = sa.Table(
            'article_translation', self.metadata,
            sa.Column('id', sa.Integer, sa.ForeignKey('article.id',
                                                      ondelete="CASCADE"),
                      primary_key=True,
                      nullable=False),
            sa.Column('locale', sa.types.CHAR(2),
                      primary_key=True,
                      nullable=False),
            sa.Column('name', sa.Unicode(255), nullable=False),
            sa.Column('content', sa.UnicodeText, nullable=False),
            sa.Column('content2', sa.UnicodeText, nullable=False))

    def create_models(self):
        class Article(ClassicBase, Translatable):
            __translatable__ = {
                'locales': self.locales,
                'auto_created_locales': True,
            }
            __translated_columns__ = [
                sa.Column('name', sa.Unicode(255)),
                sa.Column('content', sa.UnicodeText),
                sa.Column('content2', sa.UnicodeText),
            ]

            @hybrid_property
            def locale(self):
                return 'en'

        class ExtendedArticle(Article):
            pass

        class ArticleTranslation(ClassicBase, BaseTranslationMixin):
            __parent_class__ = Article

        self.Article = Article
        self.ExtendedArticle = ExtendedArticle
        self.ArticleTranslation = ArticleTranslation

    def create_mappers(self):
        sa.orm.mapper(self.Article, self.article,
                      polymorphic_on=self.article.c.discriminator)
        sa.orm.mapper(self.ExtendedArticle, self.extended_article,
                      inherits=self.Article,
                      polymorphic_identity=u'extended')
        sa.orm.mapper(self.ArticleTranslation, self.article_translation)



# -*- coding: utf-8 -*-

import sqlalchemy as sa
from sqlalchemy.ext.hybrid import hybrid_property

from sqlalchemy_i18n import Translatable, translation_base
from sqlalchemy_i18n.manager import BaseTranslationMixin
from tests import ClassicBase, ClassicTestCase, DeclarativeTestCase


class Suite(object):
    def test_converts_locale_object_to_unicode(self):
        article = self.Article()
        article.name = u'Some article'
        assert article.name == u'Some article'


class Locale(object):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return self.value

    def __unicode__(self):
        return self.value


class TestDeclarative(Suite, DeclarativeTestCase):
    def create_models(self):
        class Article(self.Model, Translatable):
            __tablename__ = 'article'

            @hybrid_property
            def locale(self):
                return Locale('en')

            id = sa.Column(sa.Integer, autoincrement=True, primary_key=True)
            description = sa.Column(sa.UnicodeText)

        class ArticleTranslation(translation_base(Article)):
            __tablename__ = 'article_translation'

            name = sa.Column(sa.Unicode(255))
            content = sa.Column(sa.UnicodeText)

        self.Article = Article


class TestClassic(Suite, ClassicTestCase):
    def create_models(self):
        class Article(ClassicBase, Translatable):
            __translatable__ = {
                'locales': self.locales,
                'default_locale': 'en',
            }
            __translated_columns__ = [
                sa.Column('name', sa.Unicode(255)),
                sa.Column('content', sa.UnicodeText),
            ]

            @hybrid_property
            def locale(self):
                return Locale('en')

        class ArticleTranslation(ClassicBase, BaseTranslationMixin):
            __parent_class__ = Article

        self.Article = Article
        self.ArticleTranslation = ArticleTranslation
